"""Dataclasses containing subgroups of config data."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class HocPaths:
    """Contains paths relative to hoc files creation."""

    hoc_dir: str
    cell_hoc_filename: str
    simul_hoc_filename: str
    run_hoc_filename: str
    syn_dir: str
    syn_dir_for_hoc: str
    syn_hoc_filename: str
    main_protocol_filename: str


@dataclass(frozen=True)
class ProtArgs:
    """Contains data needed to create protocols."""

    emodel: str
    apical_point_isec: int
    mtype: str
    prot_path: str
    features_path: str


@dataclass(frozen=True)
class SynMechArgs:
    """Contains data needed to create synapse mechanimsms.

    Attributes can be accessed only if add_synapses is True.
    """

    add_synapses: bool
    seed: int
    rng_settings_mode: str
    syn_conf_file: str
    syn_data_file: str
    syn_dir: str

    def __getattribute__(self, item):
        """Modified getattribute to restrict access to when add_synapses is True.

        Raises:
            AttributeError when an attribute other than add_synapses is looked for and
                add_synapses is False
        """
        if item != "add_synapses" and not self.add_synapses:
            raise AttributeError(
                f"You can not access {item} if add_synapses is {self.add_synapses}."
            )
        return super().__getattribute__(item)


@dataclass(frozen=True)
class MorphArgs:
    """Contains data relative to morphology."""

    morph_path: str
    do_replace_axon: bool
    axon_hoc_path: Optional[str] = None


@dataclass(frozen=True)
class PresynStimArgs:
    """Contains data relative to the presynaptic cell stimuli."""

    stim_train: float
    amp: float
    width: float



"""Feature-related classes."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

from bluepyopt.ephys.efeatures import eFELFeature

logger = logging.getLogger(__name__)


def get_feature(
    feature_config,
    main_protocol,
    protocol_name,
    recording_name,
    prefix,
):
    """Return feature name and eFelFeature.

    Args:
        feature_config (dict): contains the feature-related config data
        main_protocol (ephys.protocols.Protocol): Main Protocol containing all the protocols
        protocol_name (str): name of the protocol used
        recording_name (str): name of the recording. used to get the trace
        prefix (str): prefix used in naming responses, features, recordings, etc.

    Returns:
        (str, bluepyopt.ephys.efeatures.eFELFeature): feature name, feature
    """
    # pylint: disable=too-many-locals
    efel_feature_name = feature_config["feature"]
    meanstd = feature_config["val"]

    if hasattr(main_protocol, "subprotocols"):
        protocol = main_protocol.subprotocols()[protocol_name]
    else:
        protocol = main_protocol[protocol_name]

    feature_name = f"{prefix}.{protocol_name}.{recording_name}.{efel_feature_name}"
    recording_names = {"": f"{prefix}.{protocol_name}.{recording_name}"}

    if "strict_stim" in feature_config:
        strict_stim = feature_config["strict_stim"]
    else:
        strict_stim = True

    if hasattr(protocol, "stim_start"):
        stim_start = protocol.stim_start

        if "threshold" in feature_config:
            threshold = feature_config["threshold"]
        else:
            threshold = -30

        if "bAP" in protocol_name:
            # bAP response can be after stimulus
            stim_end = protocol.total_duration
        elif "H40S8" in protocol_name:
            stim_end = protocol.stim_last_start
        else:
            stim_end = protocol.stim_end

        stimulus_current = protocol.step_amplitude

    else:
        stim_start = None
        stim_end = None
        stimulus_current = None
        threshold = None

    feature = eFELFeature(
        feature_name,
        efel_feature_name=efel_feature_name,
        recording_names=recording_names,
        stim_start=stim_start,
        stim_end=stim_end,
        exp_mean=meanstd[0],
        exp_std=meanstd[1],
        stimulus_current=stimulus_current,
        threshold=threshold,
        int_settings={"strict_stiminterval": strict_stim},
    )

    return feature_name, feature


def define_efeatures(main_protocol, features_path, prefix=""):
    """Define the efeatures.

    Args:
        main_protocol (ephys.protocols.Protocol): Main Protocol containing all the protocols
        features_path (str): path to features file
        prefix (str): prefix used in naming responses, features, recordings, etc.

    Returns:
        dict: efeatures
    """
    with open(features_path, "r", encoding="utf-8") as features_file:
        feature_definitions = json.load(features_file)

    if "__comment" in feature_definitions:
        del feature_definitions["__comment"]

    efeatures = {}

    for protocol_name, locations in feature_definitions.items():
        for recording_name, feature_configs in locations.items():
            for feature_config in feature_configs:
                feature_name, feature = get_feature(
                    feature_config,
                    main_protocol,
                    protocol_name,
                    recording_name,
                    prefix,
                )
                efeatures[feature_name] = feature

    return efeatures



"""Functionality to fetch the experimental features and parameters from the json files."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from emodelrunner.factsheets.units import feature_units

logger = logging.getLogger(__name__)


def morphology_used_in_fitting(optimized_params_dict, emodel):
    """Returns the morphology name from finals.json used in model fitting.

    Args:
        optimized_params_dict (dict): contains the optimized parameters,
            as well as the original morphology path
        emodel (str): name of the emodel

    Returns:
        str: the original morphology name used in model fitting
    """
    emodel_params = optimized_params_dict[emodel]
    morph_path = emodel_params["morph_path"]
    morph_name = morph_path.split("/")[-1]
    return morph_name


def get_feature_dict(feature, morphology_prefix, stimulus, location, fitness):
    """Return dict containing one feature.

    Args:
        feature (dict): contains feature name and mean and std of feature
        morphology_prefix (str): prefix used in the fitness key to the feature
        stimulus (str): name of the stimulus used for this feature
        location (str): name of the location for which the feature is calculated
        fitness (dict): contains the fitness of the feature

    Returns:
        dict containing name, values, unit and model fitness of the feature
    """
    feature_name = feature["feature"]
    mean = feature["val"][0]
    std = feature["val"][1]

    try:
        unit = feature_units[feature_name]
    except KeyError:
        logger.warning(
            "%s was not found in units file. Setting unit to ''.", feature_name
        )
        unit = ""

    key_fitness = ".".join((morphology_prefix, stimulus, location, feature_name))
    try:
        fit = fitness[key_fitness]
    except KeyError:
        logger.warning(
            "%s was not found in fitness dict. Setting fitness model fitness value to ''.",
            key_fitness,
        )
        fit = ""

    return {
        "name": feature_name,
        "values": [{"mean": mean, "std": std}],
        "unit": unit,
        "model fitness": fit,
    }


def get_exp_features_data(
    emodel, morphology_prefix, features_dict, optimized_params_dict
):
    """Returns a dict containing mean and std of experimental features and model fitness.

    Args:
        emodel (str): name of the emodel
        morphology_prefix (str): prefix used in the fitness key to the feature
        features_dict (dict): contains the features
        optimized_params_dict (dict): contains the optimized parameters,
            as well as the original morphology path

    Returns:
        dict containing the output feature dicts and the original morph name used in model fitting
    """
    # pylint: disable=too-many-locals
    # it is hard to reduce number of locals without reducing readibility
    fitness = optimized_params_dict[emodel]["fitness"]

    values_dict = {}
    for stimulus, stim_data in features_dict.items():
        stim_dict = {}
        for location, loc_data in stim_data.items():
            features_list = []
            for feature in loc_data:
                features_list.append(
                    get_feature_dict(
                        feature, morphology_prefix, stimulus, location, fitness
                    )
                )
            loc_dict = {"features": features_list}
            stim_dict[location] = loc_dict
        values_dict[stimulus] = stim_dict
    values = [values_dict]

    fitted_morphology_name = morphology_used_in_fitting(optimized_params_dict, emodel)

    return {
        "name": "Experimental features",
        "values": values,
        "morphology": fitted_morphology_name,
    }



"""Functions to create cells."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from emodelrunner.cell import CellModelCustom
from emodelrunner.load import (
    load_mechanisms,
    load_syn_mechs,
    load_unoptimized_parameters,
)
from emodelrunner.morphology import create_morphology
from emodelrunner.configuration import PackageType


def create_cell(
    unopt_params_path,
    emodel,
    add_synapses,
    morph,
    gid,
    syn_mech_args=None,
    use_glu_synapse=False,
    fixhp=False,
    syn_setup_params=None,
    v_init=-80,
    celsius=34,
):
    """Create a cell.

    Args:
        unopt_params_path (str): path to the unoptimized parameters json file
        emodel (str): name to give to the cell model
        add_synapses (bool): whether to add synapses to the cell
        morph (ephys.morphologies.NrnFileMorphology): morphology object
        gid (int): cell model ID
        syn_mech_args (SynMechArgs): synapse-related configuration
        use_glu_synapse (bool): whether to use GluSynapseCustom class for synapses
        fixhp (bool): to uninsert SK_E2 for hyperpolarization in cell model
        syn_setup_params (dict): contains extra parameters to setup synapses
            when using GluSynapseCustom
        v_init (int): initial voltage (mV)
        celsius (int): cell temperature (celsius)

    Returns:
        CellModelCustom: cell model
    """
    # pylint: disable=too-many-arguments, too-many-locals
    # load mechanisms
    mechs = load_mechanisms(unopt_params_path)

    # add synapses mechs
    if add_synapses:
        mechs += [
            load_syn_mechs(
                syn_mech_args.seed,
                syn_mech_args.rng_settings_mode,
                os.path.join(syn_mech_args.syn_dir, syn_mech_args.syn_data_file),
                os.path.join(syn_mech_args.syn_dir, syn_mech_args.syn_conf_file),
                use_glu_synapse=use_glu_synapse,
                syn_setup_params=syn_setup_params,
            )
        ]

    # load parameters
    params = load_unoptimized_parameters(unopt_params_path, v_init, celsius)

    # create cell
    cell = CellModelCustom(
        name=emodel,
        morph=morph,
        mechs=mechs,
        params=params,
        gid=gid,
        add_synapses=add_synapses,
        fixhp=fixhp,
    )

    return cell


def create_cell_using_config(config):
    """Create a cell given configuration.

    Args:
        config (configparser.ConfigParser): configuration

    Raises:
        ValueError: raised when package_type is not supported

    Returns:
        CellModelCustom: cell model
    """
    unopt_params_path = config.get("Paths", "unoptimized_params_path")

    # get synapse config data
    add_synapses = config.getboolean("Synapses", "add_synapses")
    syn_mech_args = config.syn_mech_args()

    # get morphology config data
    if config.package_type in [PackageType.sscx, PackageType.thalamus]:
        morph = create_morphology(config.morph_args(), config.package_type)
    else:
        raise ValueError(f"unsupported package type: {config.package_type}")

    return create_cell(
        unopt_params_path,
        config.get("Cell", "emodel"),
        add_synapses,
        morph,
        config.getint("Cell", "gid"),
        syn_mech_args,
        v_init=config.getfloat("Cell", "v_init"),
        celsius=config.getfloat("Cell", "celsius"),
    )


def get_postcell(
    config,
    fixhp=False,
    syn_setup_params=None,
):
    """Return the postcell for synapse plasticity run.

    Args:
        config (configparser.ConfigParser): configuration
        fixhp (bool): to uninsert SK_E2 for hyperpolarization in cell model
        syn_setup_params (dict): contains extra parameters to setup synapses
            when using GluSynapseCustom

    Returns:
        CellModelCustom: post-synaptic cell model
    """
    emodel = config.get("Cell", "emodel")
    gid = config.getint("Cell", "gid")
    base_seed = config.getint("SynapsePlasticity", "base_seed")
    v_init = config.getfloat("Cell", "v_init")
    celsius = config.getfloat("Cell", "celsius")

    unopt_params_path = config.get("Paths", "unoptimized_params_path")

    # rewrite seed and rng setting mode over basic emodelrunner config defaults
    syn_mech_args = config.syn_mech_args(
        add_synapses=True, seed=base_seed, rng_settings_mode="Compatibility"
    )

    morph = create_morphology(config.synplas_morph_args(), config.package_type)

    add_synapses = True

    cell = create_cell(
        unopt_params_path,
        emodel,
        add_synapses,
        morph,
        gid,
        syn_mech_args,
        use_glu_synapse=True,
        fixhp=fixhp,
        syn_setup_params=syn_setup_params,
        v_init=v_init,
        celsius=celsius,
    )
    return cell


def get_precell(
    config,
    fixhp=False,
):
    """Return the precell for synapse plasticity pair simulation run.

     Args:
        config (configparser.ConfigParser): configuration
        fixhp (bool): to uninsert SK_E2 for hyperpolarization in cell model

    Returns:
        CellModelCustom: pre-synaptic cell model
    """
    emodel = config.get("Cell", "emodel")
    gid = config.getint("Cell", "gid")
    v_init = config.getfloat("Cell", "v_init")
    celsius = config.getfloat("Cell", "celsius")

    unopt_params_path = config.get("Paths", "precell_unoptimized_params_path")

    morph = create_morphology(
        config.synplas_morph_args(precell=True), config.package_type
    )

    add_synapses = False

    cell = create_cell(
        unopt_params_path,
        emodel,
        add_synapses,
        morph,
        gid,
        fixhp=fixhp,
        v_init=v_init,
        celsius=celsius,
    )
    return cell



"""Ephys protocols for the Thalamus packages."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-lines

import collections
import copy
import logging
import numpy as np

from bluepyopt import ephys
from tqdm import tqdm

from emodelrunner.protocols.protocols_func import CurrentOutputKeyMixin

logger = logging.getLogger(__name__)


class RatSSCxMainProtocol(ephys.protocols.Protocol):
    """Main protocol to fit RatSSCx neuron ephys parameters.

    Pseudo code:
        Find resting membrane potential
        Find input resistance
        If both of these scores are within bounds, run other protocols:
        - Find holding current
        - Find rheobase
        - Run IDRest
        - Possibly run other protocols (based on constructor arguments)
        - Return all the responses
        Otherwise return Rin and RMP protocol responses
    """

    def __init__(
        self,
        name,
        rmp_protocol=None,
        rmp_efeature=None,
        rinhold_protocol_dep=None,
        rinhold_protocol_hyp=None,
        rin_efeature_dep=None,
        rin_efeature_hyp=None,
        thdetect_protocol_dep=None,
        thdetect_protocol_hyp=None,
        other_protocols=None,
        pre_protocols=None,
    ):
        """Constructor."""
        # pylint: disable=too-many-arguments
        super().__init__(name=name)

        self.rmp_protocol = rmp_protocol
        self.rmp_efeature = rmp_efeature

        if rinhold_protocol_dep is not None:
            self.rinhold_protocol_dep = rinhold_protocol_dep
            self.rin_efeature_dep = rin_efeature_dep

        self.rinhold_protocol_hyp = rinhold_protocol_hyp
        self.rin_efeature_hyp = rin_efeature_hyp

        if thdetect_protocol_dep is not None:
            self.thdetect_protocol_dep = thdetect_protocol_dep

        self.thdetect_protocol_hyp = thdetect_protocol_hyp

        self.other_protocols = other_protocols

        self.pre_protocols = pre_protocols

    def subprotocols(self):
        """Return all the subprotocols contained in this protocol, is recursive."""
        subprotocols = collections.OrderedDict({self.name: self})
        subprotocols.update(self.rmp_protocol.subprotocols())
        if hasattr(self, "rinhold_protocol_dep"):
            subprotocols.update(self.rinhold_protocol_dep.subprotocols())
        subprotocols.update(self.rinhold_protocol_hyp.subprotocols())
        if hasattr(self, "thdetect_protocol_dep"):
            subprotocols.update(self.thdetect_protocol_dep.subprotocols())
        subprotocols.update(self.thdetect_protocol_hyp.subprotocols())
        for other_protocol in self.other_protocols:
            subprotocols.update(other_protocol.subprotocols())
        for pre_protocol in self.pre_protocols:
            subprotocols.update(pre_protocol.subprotocols())

        return subprotocols

    @property
    def rin_efeature_dep(self):
        """Get in_efeature."""
        return self.rinhold_protocol_dep.rin_efeature_dep

    @rin_efeature_dep.setter
    def rin_efeature_dep(self, value):
        """Set rin_efeature."""
        self.rinhold_protocol_dep.rin_efeature_dep = value

    @property
    def rin_efeature_hyp(self):
        """Get in_efeature."""
        return self.rinhold_protocol_hyp.rin_efeature_hyp

    @rin_efeature_hyp.setter
    def rin_efeature_hyp(self, value):
        """Set rin_efeature."""
        self.rinhold_protocol_hyp.rin_efeature_hyp = value

    def run(self, cell_model, param_values, sim=None, isolate=None):
        """Run protocol."""
        # pylint: disable=unused-argument
        responses = collections.OrderedDict()

        cell_model.freeze(param_values)

        # Find resting membrane potential
        logger.info("Running RMP protocol")
        rmp_response = self.rmp_protocol.run(cell_model, {}, sim=sim)
        responses.update(rmp_response)
        rmp = self.rmp_efeature.calculate_feature(rmp_response)

        # Find Rin and holding current
        if hasattr(self, "rinhold_protocol_dep"):
            logger.info("Running Rinhold protocol (depolarizing)")
            rinhold_response_dep = self.rinhold_protocol_dep.run(
                cell_model, {}, sim=sim, rmp=rmp
            )
            holding_current_dep = cell_model.holding_current_dep

            rin_dep = self.rin_efeature_dep.calculate_feature(rinhold_response_dep)
            responses.update(rinhold_response_dep)

        logger.info("Running Rinhold protocol (hyperpolarizing)")
        rinhold_response_hyp = self.rinhold_protocol_hyp.run(
            cell_model, {}, sim=sim, rmp=rmp
        )

        if rinhold_response_hyp is not None:
            rin_hyp = self.rin_efeature_hyp.calculate_feature(rinhold_response_hyp)

            responses.update(rinhold_response_hyp)

            if hasattr(self, "thdetect_protocol_dep"):
                logger.info("Running threshold detection protocol (depolarizing)")
                responses.update(
                    self.thdetect_protocol_dep.run(
                        cell_model,
                        {},
                        sim=sim,
                        holdi=holding_current_dep,
                        rmp=rmp,
                        rin=rin_dep,
                    )
                )

            logger.info("Running threshold detection protocol (hyperpolarizing)")
            responses.update(
                self.thdetect_protocol_hyp.run(
                    cell_model,
                    {},
                    sim=sim,
                    holdi=cell_model.holding_current_hyp,
                    rmp=rmp,
                    rin=rin_hyp,
                )
            )

            if cell_model.threshold_current_hyp is not None:
                self._run_pre_protocols(cell_model, sim, responses)
                self._run_other_protocols(cell_model, sim, responses)

        cell_model.unfreeze(param_values.keys())

        return responses

    def _run_pre_protocols(self, cell_model, sim, responses):
        """Runs the pre_protocols and updates responses dict."""
        logger.info("Running pre-protocols")
        for pre_protocol in tqdm(self.pre_protocols):
            response = pre_protocol.run(cell_model, {}, sim=sim)
            responses.update(response)

    def _run_other_protocols(self, cell_model, sim, responses):
        """Runs the other_protocols and updates responses dict."""
        logger.info("Running other protocols")
        for other_protocol in tqdm(self.other_protocols):
            response = other_protocol.run(cell_model, {}, sim=sim)
            responses.update(response)

    def generate_current(
        self, thres_i_hyp, thres_i_dep, holding_i_hyp, holding_i_dep, dt
    ):
        """Generate current for all protocols except rin and threshold detection.

        Args:
            thres_i_hyp (float): hyperpolarization threshold current (nA)
            thres_i_dep (float): depolarization threshold current (nA)
            holding_i_hyp (float): hyperpolarization holding current (nA)
            holding_i_dep (float): depolarization holding current (nA)
            dt (float): timestep of the generated currents (ms)

        Returns:
            dict containing the generated currents
        """
        currents = {}

        # rmp is step protocol
        currents.update(
            self.rmp_protocol.generate_current(
                threshold_current=thres_i_dep,
                holding_current=holding_i_dep,
                dt=dt,
            )
        )

        if self.rinhold_protocol_dep is not None:
            currents.update(
                self.rinhold_protocol_dep.generate_current(
                    threshold_current=thres_i_dep, holding_current=holding_i_dep, dt=dt
                )
            )

        currents.update(
            self.rinhold_protocol_hyp.generate_current(
                threshold_current=thres_i_hyp, holding_current=holding_i_hyp, dt=dt
            )
        )

        for pre_protocol in self.pre_protocols:
            if pre_protocol.name.endswith("_hyp"):
                pre_prot_threshold_current = thres_i_hyp
                pre_prot_holding_current = holding_i_hyp
            else:
                pre_prot_threshold_current = thres_i_dep
                pre_prot_holding_current = holding_i_dep
            currents.update(
                pre_protocol.generate_current(
                    threshold_current=pre_prot_threshold_current,
                    holding_current=pre_prot_holding_current,
                    dt=dt,
                )
            )

        for other_protocol in self.other_protocols:
            if other_protocol.name.endswith("_hyp"):
                other_prot_threshold_current = thres_i_hyp
                other_prot_holding_current = holding_i_hyp
            else:
                other_prot_threshold_current = thres_i_dep
                other_prot_holding_current = holding_i_dep
            currents.update(
                other_protocol.generate_current(
                    threshold_current=other_prot_threshold_current,
                    holding_current=other_prot_holding_current,
                    dt=dt,
                )
            )

        return currents


class RatSSCxRinHoldcurrentProtocol(ephys.protocols.Protocol):
    """IDRest protocol to fit RatSSCx neuron ephys parameters."""

    def __init__(
        self,
        name,
        rin_protocol_template=None,
        voltagebase_efeature=None,
        holdi_estimate_multiplier=2,
        holdi_precision=0.1,
        holdi_max_depth=5,
        prefix=None,
    ):
        """Constructor."""
        # pylint: disable=too-many-arguments
        super().__init__(name=name)
        self.rin_protocol_template = rin_protocol_template
        self.voltagebase_efeature = voltagebase_efeature
        self.holdi_estimate_multiplier = holdi_estimate_multiplier
        self.holdi_precision = holdi_precision
        self.holdi_max_depth = holdi_max_depth

        self.rin_efeature_dep = None
        self.rin_efeature_hyp = None

        self.prefix = "" if prefix is None else prefix + "."
        # This will be set after the run()
        self.rin_protocol = None

    def run(self, cell_model, param_values, sim, rmp=None):
        """Run protocol."""
        # Calculate Rin without holding current
        if self.name.endswith("_dep"):
            rin_noholding_protocol = self.create_rin_protocol_dep(holdi=0)
            rin_noholding_response = rin_noholding_protocol.run(
                cell_model, param_values, sim=sim
            )
            rin_noholding = self.rin_efeature_dep.calculate_feature(
                rin_noholding_response
            )

        elif self.name.endswith("_hyp"):
            rin_noholding_protocol = self.create_rin_protocol_hyp(holdi=0)
            rin_noholding_response = rin_noholding_protocol.run(
                cell_model, param_values, sim=sim
            )
            rin_noholding = self.rin_efeature_hyp.calculate_feature(
                rin_noholding_response
            )

        logger.info("Rin without holdi is %s", rin_noholding)

        logger.info(
            "Searching holdi for vhold = %s", self.voltagebase_efeature.exp_mean
        )
        # Search holding current
        holdi = self.search_holdi(
            cell_model,
            param_values,
            sim,
            self.voltagebase_efeature.exp_mean,
            rin_noholding,
            rmp,
        )

        if holdi is None:
            return None

        # Set up Rin protocol
        if self.name.endswith("_dep"):
            self.rin_protocol = self.create_rin_protocol_dep(holdi=holdi)
        elif self.name.endswith("_hyp"):
            self.rin_protocol = self.create_rin_protocol_hyp(holdi=holdi)

        # Return response
        responses = self.rin_protocol.run(cell_model, param_values, sim)

        if self.name.endswith("_dep"):
            responses[self.prefix + "bpo_holding_current_dep"] = holdi
            cell_model.holding_current_dep = holdi

        elif self.name.endswith("_hyp"):
            responses[self.prefix + "bpo_holding_current_hyp"] = holdi
            cell_model.holding_current_hyp = holdi

        return responses

    def subprotocols(self):
        """Return subprotocols."""
        subprotocols = collections.OrderedDict({self.name: self})

        subprotocols.update(
            {self.rin_protocol_template.name: self.rin_protocol_template}
        )

        return subprotocols

    def create_rin_protocol_dep(self, holdi=None):
        """Create Rin_dep protocol."""
        return self._create_rin_protocol("Rin_dep", holdi)

    def create_rin_protocol_hyp(self, holdi=None):
        """Create rin_hyp protocol."""
        return self._create_rin_protocol("Rin_hyp", holdi)

    def _create_rin_protocol(self, protocol_name, holdi):
        """Create rin protocol from self.rin_protocol_template."""
        rin_protocol = copy.deepcopy(self.rin_protocol_template)
        rin_protocol.name = protocol_name
        for recording in rin_protocol.recordings:
            recording.name = recording.name.replace(
                self.rin_protocol_template.name, rin_protocol.name
            )

        rin_protocol.holding_stimulus.step_amplitude = holdi
        return rin_protocol

    def search_holdi(
        self, cell_model, param_values, sim, holding_voltage, rin_noholding, rmp
    ):
        """Find the holding current to hold cell at holding_voltage."""
        holdi_estimate = float(holding_voltage - rmp) / rin_noholding
        logger.info(
            "Holdi estimate is %s with target vhold %s, rmp %s, Rin %s",
            holdi_estimate,
            holding_voltage,
            rmp,
            rin_noholding,
        )

        return self.binsearch_holdi(
            holding_voltage,
            cell_model,
            param_values,
            sim,
            upper_bound=0.22,
            lower_bound=self.holdi_estimate_multiplier * holdi_estimate,
            precision=self.holdi_precision,
            max_depth=self.holdi_max_depth,
        )

    def binsearch_holdi(
        self,
        holding_voltage,
        cell_model,
        param_values,
        sim=None,
        lower_bound=None,
        upper_bound=None,
        precision=None,
        max_depth=None,
        depth=1,
    ):
        """Do binary search to find holding current."""
        # pylint: disable=too-many-arguments
        middle_bound = upper_bound - abs(upper_bound - lower_bound) / 2

        if depth > max_depth:
            logger.info(
                "Search holdi reached max depth, returning with ihold %s", middle_bound
            )
            return middle_bound
        else:
            middle_voltage = self.voltage_base(
                middle_bound, cell_model, param_values, sim=sim
            )

            if middle_voltage is None:
                logger.warning("Holdi current search failed")
                return None

            if abs(middle_voltage - holding_voltage) < precision:
                logger.info(
                    "Holdi search reached precision of %s, returning with ihold %s",
                    precision,
                    middle_bound,
                )
                return middle_bound

            elif middle_voltage > holding_voltage:
                return self.binsearch_holdi(
                    holding_voltage,
                    cell_model,
                    param_values,
                    sim=sim,
                    lower_bound=lower_bound,
                    upper_bound=middle_bound,
                    precision=precision,
                    max_depth=max_depth,
                    depth=depth + 1,
                )
            elif middle_voltage < holding_voltage:
                return self.binsearch_holdi(
                    holding_voltage,
                    cell_model,
                    param_values,
                    sim=sim,
                    lower_bound=middle_bound,
                    upper_bound=upper_bound,
                    precision=precision,
                    max_depth=max_depth,
                    depth=depth + 1,
                )
            else:
                return None

    def voltage_base(self, current, cell_model, param_values, sim=None):
        """Calculate voltage base for certain stimulus current."""
        if self.name.endswith("_dep"):
            protocol = self.create_rin_protocol_dep(holdi=current)

            response = protocol.run(cell_model, param_values, sim=sim)

            feature = ephys.efeatures.eFELFeature(
                name="Holding_dep.voltage_base",
                efel_feature_name="voltage_base",
                recording_names={"": self.prefix + "Rin_dep.soma.v"},
                stim_start=protocol.step_delay,
                stim_end=protocol.step_delay + protocol.step_duration,
                exp_mean=0,
                exp_std=0.1,
            )

            voltage_base = feature.calculate_feature(response)

        elif self.name.endswith("_hyp"):
            protocol = self.create_rin_protocol_hyp(holdi=current)

            response = protocol.run(cell_model, param_values, sim=sim)

            feature = ephys.efeatures.eFELFeature(
                name="Holding_hyp.voltage_base",
                efel_feature_name="voltage_base",
                recording_names={"": self.prefix + "Rin_hyp.soma.v"},
                stim_start=protocol.step_delay,
                stim_end=protocol.step_delay + protocol.step_duration,
                exp_mean=0,
                exp_std=0.1,
            )

            voltage_base = feature.calculate_feature(response)
        return voltage_base

    def generate_current(self, threshold_current=None, holding_current=None, dt=0.1):
        """Return current time series.

        Args:
            threshold_current (float): the threshold current (nA)
            holding_current (float): the holding current (nA)
            dt (float): timestep of the generated currents (ms)

        Returns:
            dict containing the generated current
        """
        # pylint: disable=unused-argument
        holding_current = 0.0
        if self.rin_protocol.holding_stimulus is not None:
            holding_current = self.rin_protocol.holding_stimulus.step_amplitude
        total_duration = self.rin_protocol.step_stimulus.total_duration

        t = np.arange(0.0, total_duration + dt, dt)
        current = np.full(t.shape, holding_current, dtype="float64")

        ton = self.rin_protocol.step_stimulus.step_delay
        ton_idx = int(ton / dt)

        toff = (
            self.rin_protocol.step_stimulus.step_delay
            + self.rin_protocol.step_stimulus.step_duration
        )
        toff_idx = int(toff / dt)

        current[ton_idx:toff_idx] += self.rin_protocol.step_stimulus.step_amplitude

        return {self.rin_protocol.curr_output_key(): {"time": t, "current": current}}


class RatSSCxThresholdDetectionProtocol(ephys.protocols.Protocol):
    """IDRest protocol to fit RatSSCx neuron ephys parameters."""

    def __init__(
        self,
        name,
        step_protocol_template=None,
        max_threshold_voltage=None,
        holding_voltage=None,
        prefix=None,
    ):
        """Constructor."""
        super().__init__(name=name)

        self.step_protocol_template = step_protocol_template

        if max_threshold_voltage is None:
            if self.name.endswith("_dep"):
                max_threshold_voltage = -40
            elif self.name.endswith("_hyp"):
                max_threshold_voltage = -50

        self.max_threshold_voltage = max_threshold_voltage

        self.short_perc = 0.1
        self.short_steps = 5
        self.holding_voltage = holding_voltage

        self.prefix = "" if prefix is None else prefix + "."

    def subprotocols(self):
        """Return subprotocols."""
        subprotocols = collections.OrderedDict({self.name: self})

        subprotocols.update(self.step_protocol_template.subprotocols())

        return subprotocols

    def run(self, cell_model, param_values, sim, holdi, rin, rmp):
        """Run protocol."""
        # pylint: disable=unused-argument
        responses = collections.OrderedDict()

        # Calculate max threshold current
        max_threshold_current = self.search_max_threshold_current(rin=rin)

        # Calculate spike threshold
        threshold_current = self.search_spike_threshold(
            cell_model,
            {},
            holdi=holdi,
            lower_bound=0,
            upper_bound=max_threshold_current,
            sim=sim,
        )

        if self.name.endswith("_dep"):
            cell_model.threshold_current_dep = threshold_current
            responses[self.prefix + "bpo_threshold_current_dep"] = threshold_current

        if self.name.endswith("_hyp"):
            cell_model.threshold_current_hyp = threshold_current
            responses[self.prefix + "bpo_threshold_current_hyp"] = threshold_current

        return responses

    def search_max_threshold_current(self, rin=None):
        """Find the current necessary to get to max_threshold_voltage."""
        max_threshold_current = (
            float(self.max_threshold_voltage - self.holding_voltage) / rin
        )

        logger.info(
            "Max threshold current from vhold %s: %s",
            self.holding_voltage,
            max_threshold_current,
        )

        return max_threshold_current

    def create_step_protocol(self, holdi=0.0, step_current=0.0):
        """Create threshold protocol."""
        threshold_protocol = copy.deepcopy(self.step_protocol_template)

        if self.name.endswith("_dep"):
            threshold_protocol.name = "ThresholdDetection_dep"
        elif self.name.endswith("_hyp"):
            threshold_protocol.name = "ThresholdDetection_hyp"

        for recording in threshold_protocol.recordings:
            recording.name = recording.name.replace(
                self.step_protocol_template.name, threshold_protocol.name
            )

        if threshold_protocol.holding_stimulus is not None:
            threshold_protocol.holding_stimulus.step_amplitude = holdi
        threshold_protocol.step_stimulus.step_amplitude = step_current

        return threshold_protocol

    def create_short_threshold_protocol(self, holdi=None, step_current=None):
        """Create short threshold protocol."""
        short_protocol = self.create_step_protocol(
            holdi=holdi, step_current=step_current
        )
        origin_step_duration = short_protocol.step_stimulus.step_duration
        origin_step_delay = short_protocol.step_stimulus.step_delay

        short_step_duration = origin_step_duration * self.short_perc
        short_total_duration = origin_step_delay + short_step_duration

        short_protocol.step_stimulus.step_duration = short_step_duration
        short_protocol.step_stimulus.total_duration = short_total_duration

        if short_protocol.holding_stimulus is not None:
            short_protocol.holding_stimulus.step_duration = short_total_duration
            short_protocol.holding_stimulus.total_duration = short_total_duration

        return short_protocol

    def detect_spike(
        self,
        cell_model,
        param_values,
        sim=None,
        step_current=None,
        holdi=None,
        short=False,
    ):
        """Detect if spike is present at current level."""
        # Only run short pulse if percentage set smaller than 100%
        if short and self.short_perc < 1.0:
            protocol = self.create_short_threshold_protocol(
                holdi=holdi, step_current=step_current
            )
        else:
            protocol = self.create_step_protocol(holdi=holdi, step_current=step_current)

        response = protocol.run(cell_model, param_values, sim=sim)

        if self.name.endswith("_dep"):
            feature = ephys.efeatures.eFELFeature(
                name="ThresholdDetection_dep.Spikecount",
                efel_feature_name="Spikecount_stimint",
                recording_names={"": self.prefix + "ThresholdDetection_dep.soma.v"},
                stim_start=protocol.step_delay,
                stim_end=protocol.step_delay + protocol.step_duration,
                exp_mean=1,
                exp_std=0.1,
            )

        elif self.name.endswith("_hyp"):
            feature = ephys.efeatures.eFELFeature(
                name="ThresholdDetection_hyp.Spikecount",
                efel_feature_name="Spikecount_stimint",
                recording_names={"": self.prefix + "ThresholdDetection_hyp.soma.v"},
                stim_start=protocol.step_delay,
                stim_end=protocol.step_delay + protocol.step_duration,
                exp_mean=1,
                exp_std=0.1,
            )

        spike_count = feature.calculate_feature(response)
        logger.debug("%s spikes with I = %s", spike_count, step_current)
        return spike_count >= 1

    def binsearch_spike_threshold(
        self,
        cell_model,
        param_values,
        sim=None,
        holdi=None,
        lower_bound=None,
        upper_bound=None,
        precision=0.01,
        max_depth=5,
        depth=1,
    ):
        """Do binary search to find spike threshold.

        Assumption is that lower_bound has no spike, upper_bound has.
        """
        # pylint: disable=too-many-arguments
        if depth > max_depth or abs(upper_bound - lower_bound) < precision:
            return upper_bound
        middle_bound = upper_bound - abs(upper_bound - lower_bound) / 2
        spike_detected = self.detect_spike(
            cell_model,
            param_values,
            sim=sim,
            holdi=holdi,
            step_current=middle_bound,
            short=False,
        )
        if spike_detected:
            return self.binsearch_spike_threshold(
                cell_model,
                param_values,
                sim=sim,
                holdi=holdi,
                lower_bound=lower_bound,
                upper_bound=middle_bound,
                depth=depth + 1,
            )
        else:
            return self.binsearch_spike_threshold(
                cell_model,
                param_values,
                sim=sim,
                holdi=holdi,
                lower_bound=middle_bound,
                upper_bound=upper_bound,
                depth=depth + 1,
            )

    def search_spike_threshold(
        self,
        cell_model,
        param_values,
        sim=None,
        holdi=None,
        lower_bound=None,
        upper_bound=None,
    ):
        """Find the current step spiking threshold."""
        step_currents = np.linspace(lower_bound, upper_bound, num=self.short_steps)

        if len(step_currents) == 0:
            return None

        for step_current in step_currents:
            latest_step_current = step_current
            spike_detected = self.detect_spike(
                cell_model,
                param_values,
                sim=sim,
                holdi=holdi,
                step_current=step_current,
                short=True,
            )

            if spike_detected:
                upper_bound = step_current
                break

        # if upper bound didn't have spike with short stimulus
        # check if there is one with longer stimulus
        if not spike_detected:
            spike_detected = self.detect_spike(
                cell_model,
                param_values,
                sim=sim,
                holdi=holdi,
                step_current=latest_step_current,
                short=False,
            )

            if spike_detected:
                upper_bound = latest_step_current
            else:
                return None

        threshold_current = self.binsearch_spike_threshold(
            cell_model,
            {},
            sim=sim,
            holdi=holdi,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
        )
        logger.info("Threshold current from %s is %s", holdi, threshold_current)
        return threshold_current


class StepProtocolCustom(ephys.protocols.StepProtocol, CurrentOutputKeyMixin):
    """Step protocol with custom options to turn stochkv_det on or off."""

    def __init__(
        self,
        name=None,
        step_stimulus=None,
        holding_stimulus=None,
        recordings=None,
        cvode_active=None,
        stochkv_det=None,
    ):
        """Constructor."""
        super().__init__(
            name,
            step_stimulus=step_stimulus,
            holding_stimulus=holding_stimulus,
            recordings=recordings,
            cvode_active=cvode_active,
        )

        self.stochkv_det = stochkv_det

    def run(self, cell_model, param_values, sim=None, isolate=None, timeout=None):
        """Run protocol."""
        responses = {}

        if self.stochkv_det is not None and not self.stochkv_det:
            for mechanism in cell_model.mechanisms:
                if "Stoch" in mechanism.prefix:
                    mechanism.deterministic = False
            self.cvode_active = False

        responses.update(
            super().run(
                cell_model, param_values, sim=sim, isolate=isolate, timeout=timeout
            )
        )

        if self.stochkv_det is not None and not self.stochkv_det:
            for mechanism in cell_model.mechanisms:
                if "Stoch" in mechanism.prefix:
                    mechanism.deterministic = True
            self.cvode_active = True

        return responses

    def generate_current(self, threshold_current=None, holding_current=None, dt=0.1):
        """Return current time series.

        Args:
            threshold_current (float): the threshold current (nA)
            holding_current (float): the holding current (nA)
            dt (float): timestep of the generated currents (ms)

        Returns:
            dict containing the generated current
        """
        # pylint: disable=unused-argument
        holding_current = 0.0
        if self.holding_stimulus is not None:
            holding_current = self.holding_stimulus.step_amplitude
        total_duration = self.step_stimulus.total_duration

        t = np.arange(0.0, total_duration + dt, dt)
        current = np.full(t.shape, holding_current, dtype="float64")

        ton = self.step_stimulus.step_delay
        ton_idx = int(ton / dt)

        toff = self.step_stimulus.step_delay + self.step_stimulus.step_duration
        toff_idx = int(toff / dt)

        current[ton_idx:toff_idx] += self.step_stimulus.step_amplitude

        return {self.curr_output_key(): {"time": t, "current": current}}

    @property
    def stim_start(self):
        """Time stimulus starts.

        Returns:
            the time at which the stimulus starts (ms)
        """
        return self.step_stimulus.step_delay

    @property
    def stim_end(self):
        """Time stimulus ends.

        Returns:
            the time at which the stimulus ends (ms)
        """
        return self.step_stimulus.step_delay + self.step_stimulus.step_duration

    @property
    def step_amplitude(self):
        """Stimuli amplitude.

        Returns:
            the amplitude of the step stimuli (nA)
        """
        return self.step_stimulus.step_amplitude


class StepThresholdProtocol(StepProtocolCustom):
    """Step protocol based on threshold."""

    def __init__(
        self,
        name,
        thresh_perc=None,
        step_stimulus=None,
        holding_stimulus=None,
        recordings=None,
        cvode_active=None,
        stochkv_det=None,
    ):
        """Constructor."""
        super().__init__(
            name,
            step_stimulus=step_stimulus,
            holding_stimulus=holding_stimulus,
            recordings=recordings,
            cvode_active=cvode_active,
        )

        self.thresh_perc = thresh_perc
        self.stochkv_det = stochkv_det

    def run(self, cell_model, param_values, sim=None, isolate=None, timeout=None):
        """Run protocol."""
        logger.info("Running protocol %s", self.name)
        responses = {}
        if not hasattr(cell_model, "threshold_current_hyp"):
            raise AttributeError(
                f"StepThresholdProtocol: running on cell_model "
                f"that doesnt have threshold current value set: {str(cell_model)}",
            )

        if self.name.endswith("_hyp"):
            self.holding_stimulus.step_amplitude = cell_model.holding_current_hyp

            self.step_stimulus.step_amplitude = cell_model.threshold_current_hyp * (
                float(self.thresh_perc) / 100
            )

        else:
            self.holding_stimulus.step_amplitude = cell_model.holding_current_dep

            self.step_stimulus.step_amplitude = cell_model.threshold_current_dep * (
                float(self.thresh_perc) / 100
            )

        if self.stochkv_det is not None and not self.stochkv_det:
            for mechanism in cell_model.mechanisms:
                if "Stoch" in mechanism.prefix:
                    mechanism.deterministic = False
            self.cvode_active = False

        responses.update(
            super().run(
                cell_model, param_values, sim=sim, isolate=isolate, timeout=timeout
            )
        )

        if self.stochkv_det is not None and not self.stochkv_det:
            for mechanism in cell_model.mechanisms:
                if "Stoch" in mechanism.prefix:
                    mechanism.deterministic = True
            self.cvode_active = True

        return responses

    def generate_current(self, threshold_current, holding_current, dt=0.1):
        """Return current time series.

        Args:
            threshold_current (float): the threshold current (nA)
            holding_current (float): the holding current (nA)
            dt (float): timestep of the generated currents (ms)

        Returns:
            dict containing the generated current
        """
        # pylint: disable=signature-differs
        total_duration = self.step_stimulus.total_duration

        t = np.arange(0.0, total_duration + dt, dt)
        current = np.full(t.shape, holding_current, dtype="float64")

        ton = self.step_stimulus.step_delay
        ton_idx = int(ton / dt)

        toff = self.step_stimulus.step_delay + self.step_stimulus.step_duration
        toff_idx = int(toff / dt)

        if threshold_current is not None:
            current[ton_idx:toff_idx] += threshold_current * (
                float(self.thresh_perc) / 100.0
            )

        return {self.curr_output_key(): {"time": t, "current": current}}



"""Synapse Stimuli."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from bluepyopt import ephys


class NrnNetStimStimulusCustom(ephys.stimuli.Stimulus):
    """Current stimulus based on current amplitude and time series.

    Attributes:
        total_duration (float): duration of run (ms)
        locations (list): synapse point processes locations to connect to
        interval (float): time between spikes (ms)
        number (int): average number of spikes
        start (float): most likely start time of first spike (ms)
        noise (float): fractional randomness (0 deterministic,
                1 negexp interval distrubtion)
        connections (dict): contains simulator NetCon and NetStim
            so that they are persistent
    """

    def __init__(
        self,
        locations=None,
        total_duration=None,
        interval=None,
        number=None,
        start=None,
        noise=0,
    ):
        """Constructor.

        Args:
            locations (list): synapse point processes locations to connect to
            total_duration (float): duration of run (ms)
            interval (float): time between spikes (ms)
            number (int): average number of spikes
            start (float): most likely start time of first spike (ms)
            noise (float): fractional randomness (0 deterministic,
                   1 negexp interval distrubtion)
        """
        super().__init__()
        if total_duration is None:
            raise ValueError("NrnNetStimStimulus: Need to specify a total duration")
        self.total_duration = total_duration

        self.locations = locations
        self.interval = interval
        self.number = number
        self.start = start
        self.noise = noise
        self.connections = {}

    def instantiate(self, sim=None, icell=None):
        """Instantiate stimuli and connections.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        if self.connections is None:
            self.connections = {}
        for location in self.locations:
            self.connections[location.name] = []
            for synapse in location.instantiate(sim=sim, icell=icell):
                netstim = sim.neuron.h.NetStim()
                netstim.interval = (
                    synapse.interval if synapse.interval is not None else self.interval
                )
                netstim.number = (
                    synapse.number if synapse.number is not None else self.number
                )
                netstim.start = (
                    synapse.start if synapse.start is not None else self.start
                )
                netstim.noise = (
                    synapse.noise if synapse.noise is not None else self.noise
                )
                netcon = sim.neuron.h.NetCon(
                    netstim, synapse.hsynapse, -30, synapse.delay, synapse.weight
                )

                self.connections[location.name].append((netcon, netstim))

    def destroy(self, sim=None):
        """Destroy stimulus.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        # pylint: disable=unused-argument
        self.connections = None

    def __str__(self):
        """String representation."""
        # pylint: disable=consider-using-f-string
        return (
            "Netstim at %s" % ",".join(location for location in self.locations)
            if self.locations is not None
            else "Netstim"
        )


class NrnVecStimStimulusCustom(ephys.stimuli.Stimulus):
    """Current stimulus based on stochastic current amplitude.

    Attributes:
        total_duration (float): time after which no synapses are allowed to fire (ms)
        locations (list): synapse point processes locations to connect to
        start (float): most likely start time of first spike (ms)
        seed (int): seed for random number generator
        vecstim_random (str): origin of the random nmb gener. can be "python" or "neuron"
        pre_spike_train (list): list of spike train. If None, will be generated by random numbers
        connections (dict): contains simulator NetCon and VecStim and time Vector
            so that they are persistent
    """

    def __init__(
        self,
        locations=None,
        start=None,
        stop=None,
        seed=1,
        vecstim_random="python",
        pre_spike_train=None,
    ):
        """Constructor.

        Args:
            locations (list): synapse point processes locations to connect to
            start (float): most likely start time of first spike (ms)
            stop (float): time after which no synapses are allowed to fire (ms)
            seed (int): seed for random number generator
            vecstim_random (str): origin of the random nmb gener. can be "python" or "neuron"
            pre_spike_train (list): list of spike train.
                If None, will be generated by random numbers
        """
        super().__init__()
        if stop is None:
            raise ValueError("NrnVecStimStimulus: Need to specify a stop time")
        # must be named total_duration because of ephys.protocols
        self.total_duration = stop

        self.locations = locations
        self.start = start
        self.seed = seed
        self.vecstim_random = vecstim_random
        self.pre_spike_train = pre_spike_train
        self.connections = {}

    def instantiate(self, sim=None, icell=None):
        """Instantiate stimuli and connections.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        if self.connections is None:
            self.connections = {}

        if self.pre_spike_train is None:
            if self.vecstim_random == "python":
                random.seed(self.seed)
            else:
                rand = sim.neuron.h.Random(self.seed)
                rand.uniform(self.start, self.total_duration)

        for location in self.locations:
            self.connections[location.name] = []
            for synapse in location.instantiate(sim=sim, icell=icell):
                # create spike_train
                if self.pre_spike_train is not None:
                    spike_train = self.pre_spike_train
                else:
                    if self.vecstim_random == "python":
                        spike_train = [random.uniform(self.start, self.total_duration)]
                    else:
                        spike_train = [rand.repick()]

                t_vec = sim.neuron.h.Vector(spike_train)
                vecstim = sim.neuron.h.VecStim()
                vecstim.play(t_vec, sim.dt)
                netcon = sim.neuron.h.NetCon(
                    vecstim, synapse.hsynapse, -30, synapse.delay, synapse.weight
                )

                self.connections[location.name].append((netcon, vecstim, t_vec))

    def destroy(self, sim=None):
        """Destroy stimulus.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        # pylint: disable=unused-argument
        self.connections = None

    def __str__(self):
        """String representation."""
        # pylint: disable=consider-using-f-string
        return (
            "Vecstim at %s" % ",".join(location for location in self.locations)
            if self.locations is not None
            else "Vecstim"
        )


class NetConSpikeDetector(ephys.stimuli.Stimulus):
    """Netcon linking output from pre-cell to post-cell's synapses.

    Attributes:
        total_duration (float): end time of connection (ms)
        locations (list): synapse point processes locations to connect to
        connections (dict): contains simulator NetCon so that they are persistent
    """

    def __init__(self, total_duration=None, locations=None):
        """Constructor.

        Args:
            total_duration (float): end time of connection (ms)
            locations (list): synapse point processes locations to connect to
        """
        self.total_duration = total_duration
        self.locations = locations

        self.connections = {}

    def instantiate(self, sim, icell):
        """Instantiate connections.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        if self.connections is None:
            self.connections = {}

        for location in self.locations:
            self.connections[location.name] = []
            for synapse in location.instantiate(sim=sim, icell=icell):
                # taken from bglibpy.cell.create_netcon_spikedetector
                # M. Hines magic to return a variable by reference to a python function
                netcon = sim.neuron.h.ref(None)
                icell.getCell().connect2target(synapse.hsynapse, netcon)
                netcon = netcon[0]
                netcon.weight[0] = synapse.weight
                netcon.delay = synapse.delay

                self.connections[location.name].append(netcon)

    def destroy(self, sim=None):
        """Destroy stimulus.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        # pylint: disable=unused-argument
        self.connections = None

    def __str__(self):
        """String representation."""
        # pylint: disable=consider-using-f-string
        return (
            "NetconSpikeDetector at %s"
            % ",".join(location for location in self.locations)
            if self.locations is not None
            else "NetconSpikeDetector"
        )



"""Contains the units of the features used in factsheets."""
from types import MappingProxyType
from efel.units import _units as efel_units

feature_units = efel_units.copy()

# add the bluepyopt units
feature_units["bpo_holding_current"] = "nA"
feature_units["bpo_threshold_current"] = "nA"

# make feature_units immutable
feature_units = MappingProxyType(feature_units)



"""Custom Cell class."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from bluepyopt import ephys

from emodelrunner.create_hoc_tools import create_hoc

logger = logging.getLogger(__name__)


class CellModelCustom(ephys.models.CellModel):
    """Cell model class.

    Attributes:
        name (str): name of the model
        morphology (bluepyopt.ephys.morphologies.Morphology):
            underlying Morphology of the cell
        mechanisms (list of bluepyopt.ephys.mechanisms.Mechanisms):
            Mechanisms associated with the cell
        params (collections.OrderedDict of bluepyopt.Parameters): Parameters of the cell model
        icell (neuron cell): Cell instantiation in simulator
        param_values (dict): contains values of the optimized parameters
        gid (int): cell's ID
        seclist_names (list of strings): Names of the lists of sections
        secarray_names (list of strings): Names of the sections
        add_synapses (bool): set to True to add synapses to the cell
        fixhp (bool): to uninsert SK_E2 for hyperpolarization
    """

    def __init__(
        self,
        name,
        morph=None,
        mechs=None,
        params=None,
        gid=0,
        add_synapses=False,
        fixhp=False,
    ):
        """Constructor.

        Args:
            name (str): name of this object
                        should be alphanumeric string, underscores are allowed,
                        first char should be a letter
            morph (Morphology): underlying Morphology of the cell
            mechs (list of Mechanisms): Mechanisms associated with the cell
            params (list of Parameters): Parameters of the cell model
            gid (int): id of cell
            add_synapses (bool): set to True to add synapses to the cell
            fixhp (bool): to uninsert SK_E2 for hyperpolarization
        """
        super().__init__(name, morph, mechs, params, gid)
        self.add_synapses = add_synapses
        self.fixhp = fixhp

    def get_replace_axon(self):
        """Return appropriate replace_axon str.

        Returns:
            str: hoc script defining the replacement axon
        """
        if self.morphology.do_replace_axon:
            replace_axon = self.morphology.replace_axon_hoc
        else:
            replace_axon = None

        if (
            self.morphology.morph_modifiers is not None
            and self.morphology.morph_modifiers_hoc is None
        ):
            logger.warning(
                "You have provided custom morphology modifiers, \
                            but no corresponding hoc files."
            )
        elif (
            self.morphology.morph_modifiers is not None
            and self.morphology.morph_modifiers_hoc is not None
        ):
            if replace_axon is None:
                replace_axon = ""
            for morph_modifier_hoc in self.morphology.morph_modifiers_hoc:
                replace_axon += "\n"
                replace_axon += morph_modifier_hoc

        return replace_axon

    def freeze_params(self, param_values):
        """Freeze params and return list of params to unfreze afterwards.

        Args:
            param_values (dict): contains values of the optimized parameters

        Returns:
            list: names of the newly frozen parameters
        """
        to_unfreeze = []
        for param in self.params.values():
            if not param.frozen:
                param.freeze(param_values[param.name])
                to_unfreeze.append(param.name)

        return to_unfreeze

    def remove_point_process_mechs(self):
        """Return mechanisms without point process mechanisms.

        Returns:
            list of Mechanisms without the point process mechanisms
        """
        mechs = []
        for mech in self.mechanisms:
            if not hasattr(mech, "pprocesses"):
                mechs.append(mech)

        return mechs

    def create_custom_hoc(
        self,
        param_values,
        ignored_globals=(),
        template_path="templates/cell_template.jinja2",
        disable_banner=False,
        syn_dir=None,
        syn_hoc_filename=None,
        syn_temp_name="hoc_synapses",
    ):
        """Create hoc code for this model.

        Args:
            param_values (dict): contains values of the optimized parameters
            ignored_globals (iterable str): HOC coded is added for each
                NrnGlobalParameter that exists, to test that it matches the values
                set in the parameters.
                This iterable contains parameter names that aren't checked
            template_path (str): path to the jinja2 template
            disable_banner (bool): if not True: a banner is added to the hoc file
            syn_dir (str): directory where the synapse data /files are
            syn_hoc_filename (str): file name of synapse hoc file
            syn_temp_name (str): synapse class name in hoc

        Returns:
            str: hoc script describing this cell model
        """
        # pylint: disable=too-many-arguments
        to_unfreeze = self.freeze_params(param_values)

        replace_axon = self.get_replace_axon()

        mechs = self.remove_point_process_mechs()

        ret = create_hoc(
            mechs=mechs,
            parameters=self.params.values(),
            ignored_globals=ignored_globals,
            replace_axon=replace_axon,
            template_name=self.name,
            template_path=template_path,
            disable_banner=disable_banner,
            add_synapses=self.add_synapses,
            synapses_template_name=syn_temp_name,
            syn_hoc_filename=syn_hoc_filename,
            syn_dir=syn_dir,
        )

        self.unfreeze(to_unfreeze)

        return ret

    @staticmethod
    def connectable_empty_cell_template(
        template_name,
        seclist_names=None,
        secarray_names=None,
    ):
        """Create a hoc script of an empty cell with an additional connection function.

        Args:
            template_name (str): name of the cell model
            seclist_names (list of strings): Names of the lists of sections
            secarray_names (list of strings): Names of the sections

        Returns:
            str: hoc script describing an empty cell model with connection function
        """
        objref_str = "objref this, CellRef"
        newseclist_str = ""

        if seclist_names:
            for seclist_name in seclist_names:
                objref_str += f", {seclist_name}"
                newseclist_str += f"             {seclist_name} = new SectionList()\n"

        secarrays_str = ""
        if secarray_names:
            secarrays_str = "create "
            secarrays_str += ", ".join(
                f"{secarray_name}[1]" for secarray_name in secarray_names
            )
            secarrays_str += "\n"

        template = """\
        begintemplate %(template_name)s
          %(objref_str)s
          public init, getCell, getCCell, connect2target
          public soma, dend, apic, axon, myelin
          %(secarrays_str)s

          obfunc getCell(){
            return this
          }
          obfunc getCCell(){
            return CellRef
          }

          /*!
          * @param $o1 NetCon source (can be nil)
          * @param $o2 Variable where generated NetCon will be placed
          */
          proc connect2target() { //$o1 target point process, $o2 returned NetCon
            soma $o2 = new NetCon(&v(1), $o1)
            $o2.threshold = -30
          }

          proc init() {\n%(newseclist_str)s
            forall delete_section()
            CellRef = this
          }

          gid = 0

          proc destroy() {localobj nil
            CellRef = nil
          }


        endtemplate %(template_name)s
               """ % {
            "template_name": template_name,
            "objref_str": objref_str,
            "newseclist_str": newseclist_str,
            "secarrays_str": secarrays_str,
        }
        return template

    @staticmethod
    def create_empty_cell(name, sim, seclist_names=None, secarray_names=None):
        """Create an empty cell in Neuron.

        Args:
            template_name (str): name of the cell model
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            seclist_names (list of strings): Names of the lists of sections
            secarray_names (list of strings): Names of the sections

        Returns:
            Cell instantiation in simulator
        """
        hoc_template = CellModelCustom.connectable_empty_cell_template(
            name, seclist_names, secarray_names
        )
        sim.neuron.h(hoc_template)

        template_function = getattr(sim.neuron.h, name)

        return template_function()

    def instantiate(self, sim=None):
        """Instantiate model in simulator.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        # pylint: disable=unnecessary-comprehension
        super().instantiate(sim)

        # Hyperpolarization workaround
        somatic = [x for x in self.icell.somatic]
        axonal = [x for x in self.icell.axonal]
        if self.fixhp:
            for sec in somatic + axonal:
                sec.uninsert("SK_E2")



"""Analysis tools for synapse plasticity."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import h5py
import numpy as np


def get_epsp_vector(t, v, spikes, window):
    """Extract EPSPs at time `spikes` from voltage trace `v`.

    Args:
        t (numpy.ndarray): Time vector (ms).
        v (numpy.ndarray): Soma voltage trace. Must have the same size of `t` (mV).
        spikes (numpy.ndarray): Time of presynaptic spikes.
            Every value in `spikes` must be in the interval [min(`t`), max(`t`)] (ms).
        window (float): Size of EPS detection window in ms.

    Returns:
        numpy.ndarray: Vector of EPSPs (mV).

    Raises:
        RuntimeError: if a postsynaptic cell spike is found during a connectivity test
        RuntimeError: if the detection window can have multiple EPSPs in it
    """
    # Verify presence of only one EPSP in window
    if np.max(np.diff(spikes)) < window:
        raise RuntimeError("Only one EPSP should be present in the detection window")

    # Get EPSPs
    n = len(spikes)
    epsps = np.zeros(n)
    for i in range(n):
        w0 = np.searchsorted(t, spikes[i])  # Beginning of EPSP window
        vbaseline = v[w0]
        w1 = np.searchsorted(t, spikes[i] + window)  # End of EPSP window
        epspmaxidx = np.argmax(v[w0:w1]) + w0
        vepsp = v[epspmaxidx]
        # Abort if postsynaptic spike
        if vepsp > -30:
            raise RuntimeError("Postsynaptic cell spiking during connectivity test")
        # Compute epsps
        epsp = vepsp - vbaseline
        epsps[i] = epsp
    return epsps


def epsp_slope(vtrace):
    """Returns slope between points at [0.3, 0.75] * peak of given trace.

    Attention! This function expects that the trace is sampled (or interpolated)
    at a single timestep, since time is not taken into account during calculations.

    Args:
        vtrace (numpy.ndarray): interpolated voltage trace of one EPSP (mV)

    Returns:
        float: slope of the EPSP (mV)
    """
    # Remove baseline
    v = vtrace - vtrace[0]
    # Find peak
    peak = np.max(v)
    peak_idx = np.argmax(v)
    # Get 30% and 75% rise time indices
    idx0 = np.argmin(np.abs(v[:peak_idx] - 0.3 * peak))
    idx1 = np.argmin(np.abs(v[:peak_idx] - 0.75 * peak))
    # Get slope of fitting line between the two
    m = (v[idx1] - v[idx0]) / (idx1 - idx0)
    return m


class Experiment:
    """A full STDP induction experiment.

    The experiment consists of two connectivity tests (C01 and C02), separed
    by the induction protocol.

    Attributes:
        t (numpy.ndarray): Time vector (ms)
        v (numpy.ndarray): Soma voltage trace (ms)
        spikes (numpy.ndarray): Pre-synaptic spike times (ms)
        duration (dict): Duration of C01 and C02 phases (ms)
        period (float): The Cx period (ms)
        c01period (float): The C01 period (ms)
        c02period (float): The C02 period (ms)
        epspwindow (float): Size of EPS detection window (ms)
        cxs (list): Names of the Cx tests. These names are used as keys
            in the dictionaries (cxspikes, duration, epsp, cxtrace)
        cxspikes (dict): Pre-synaptic spikes times (ms)
            occuring during each test (C01 & C02)
    """

    def __init__(
        self,
        data,
        c01duration=40.0,
        c02duration=40.0,
        period=10.0,
        c01period=None,
        c02period=None,
    ):
        """Constructor.

        Args:
            data (str or dict): Path to simulation results file or results dictionary
            c01duration (float): Duration of C01 phase in minutes.
            c02duration (float): Duration of C02 phase in minutes.
            period (float): The Cx period in seconds.
            c01period (float): The C01 period in seconds. If None, period will be used.
            c02period (float): The C02 period in seconds. If None, period will be used.

        Raises:
            TypeError if data is neither a dictionary nor a string
        """
        if isinstance(data, dict):
            self.t = data["t"]
            self.v = data["v"]
            self.spikes = data["prespikes"]
        elif isinstance(data, str):
            h5file = h5py.File(data, "r")
            self.t = h5file["t"][()]
            self.v = h5file["v"][()]
            self.spikes = h5file["prespikes"][()]
            h5file.close()
        else:
            raise TypeError("data argument of Experiment has to be dict or str")
        # Store other attributes
        self.duration = {
            "C01": c01duration * 60 * 1000.0,  # min to ms
            "C02": c02duration * 60 * 1000.0,
        }  # min to ms
        self.period = period * 1000.0  # sec to ms
        self.c01period = (
            c01period * 1000.0 if c01period is not None else period * 1000.0
        )
        self.c02period = (
            c02period * 1000.0 if c02period is not None else period * 1000.0
        )
        # Create common attributes
        self.epspwindow = 100.0  # ms
        self.cxs = ["C01", "C02"]
        self.cxspikes = {
            "C01": self.spikes[: int(self.duration["C01"] / self.c01period)],
            "C02": self.spikes[-int(self.duration["C02"] / self.c02period) :],
        }

    @property
    def epsp(self):
        """Returns dict containing EPSP vectors for each test (mV)."""
        epsp = {
            cx: get_epsp_vector(self.t, self.v, self.cxspikes[cx], self.epspwindow)
            for cx in self.cxs
        }
        return epsp

    @property
    def cxtrace(self):
        """Returns time and interpolated trace of each EPSP as dict."""
        tdense = np.linspace(0, self.epspwindow, int(self.epspwindow / 0.025))
        cxtrace = {"t": tdense}
        for cx in self.cxs:
            cxtrace[cx] = []
            for s in self.cxspikes[cx]:
                idx0 = np.searchsorted(self.t, s)
                idx1 = np.searchsorted(self.t, s + self.epspwindow)
                vdense = np.interp(tdense + s, self.t[idx0:idx1], self.v[idx0:idx1])
                cxtrace[cx].append(vdense)
            cxtrace[cx] = np.array(cxtrace[cx])

        return cxtrace

    def compute_epsp_interval(self, interval):
        """Compute mean EPSP amplitude at regular intervals.

        Args:
            interval (float): The interval in minutes.

        Returns:
            dict: A dictionary of interval statistics for each Cx in Experiment.
        """
        results = {}
        for cx in self.cxs:
            n = int(self.duration[cx] / (interval * 60 * 1000.0))
            epsp_groups = np.split(self.epsp[cx], n)
            spike_groups = np.split(self.cxspikes[cx], n)
            avg = np.mean(epsp_groups, axis=1)
            sem = np.std(epsp_groups, axis=1) / np.sqrt(len(epsp_groups[0]))
            t = np.mean(spike_groups, axis=1)
            results[cx] = {"avg": avg, "sem": sem, "t": t}
        return results

    def compute_epsp_ratio(self, n, method="amplitude", full=False):
        """Compute mean EPSP change.

        Args:
            n (int): Number of sweeps in Cx to be considered for mean EPSP calculation
            method (str): Method used to compute EPSP ratio (amplitude or slope)
            full (bool): whether to return the mean value and std of epsp before and after
                cell stimulus in addition to the epsp ratio

        Returns:
            if full is False, returns

            - ratio_epsp (float): Mean EPSP change

            if full is True, returns a tuple containing

            - epsp_before (float): mean value of EPSP or slope of mean EPSP trace in test C01
            - epsp_after (float): mean value of EPSP or slope of mean EPSP trace in test C02
            - ratio_epsp (float): Mean EPSP change
            - epsp_before_std (float): std of EPSP in test C01 if method is amplitude, else 0
            - epsp_after_std (float): std of EPSP in test C02 if method is amplitude, else 0


        Raises:
            ValueError: if method is neither 'amplitude' nor 'slope'
        """
        if method == "amplitude":
            epsp_before = np.mean(self.epsp["C01"][-n:])
            epsp_before_std = np.std(self.epsp["C01"][-n:])
            epsp_after = np.mean(self.epsp["C02"][-n:])
            epsp_after_std = np.std(self.epsp["C02"][-n:])
        elif method == "slope":
            epsp_before = epsp_slope(np.mean(self.cxtrace["C01"][-n:], axis=0))
            epsp_before_std = 0
            epsp_after = epsp_slope(np.mean(self.cxtrace["C02"][-n:], axis=0))
            epsp_after_std = 0
        else:
            raise ValueError(f"Unknown method {method}")
        epsp_ratio = epsp_after / epsp_before
        if full:
            return epsp_before, epsp_after, epsp_ratio, epsp_before_std, epsp_after_std
        else:
            return epsp_ratio

    def normalize_time(self, t):
        """Normalize time vector `t`.

        Convert milliseconds to minutes and shift t = 0 to beginning of induction phase.

        Args:
            t (numpy.ndarray): Time vector (ms)

        Results:
            tnorm (numpy.ndarray): Normalized time vector (mn)
        """
        return (t - self.duration["C01"]) / (60.0 * 1000.0)
